(*
 *
 * Copyright (c) 2004, 
 *  Jeremy Condit       <jcondit@cs.berkeley.edu>
 *  George C. Necula    <necula@cs.berkeley.edu>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright (c) 2007 Intel Corporation 
 * All rights reserved. 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 	Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 	Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *     Neither the name of the Intel Corporation nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE INTEL OR ITS
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *)

module C = Cil

let home : string ref = ref "/usr/local/lib/heapsafe"
let outFile : string ref = ref ""
let debug : bool ref = ref false
let parseFile : string ref = ref ""
let warnTypeofChar: bool ref = ref false
let opsFile : string ref = ref "heapsafe/rcops.h"
let noRc: bool ref = ref false
let noDefer: bool ref = ref false
let noLocals: bool ref = ref false
let size_t: string ref = ref ""
let parsedFile: string ref = ref ""
let typeDebug: bool ref = ref false
let fakeAdjust: bool ref = ref false
let saveAdjust: bool ref = ref false
let adjustDir: string ref = ref ""
let noHoist: bool ref = ref false
let noMerge: bool ref = ref false


let rec options = [
  (* General *)
  "", Arg.Unit (fun () -> ()), "General:";
  "--hsdebug", Arg.Set debug,
    "Compile for debugging";
  "--warnall", Arg.Unit (fun _ -> Errormsg.warnFlag := true),
    "Show all warnings";
  "--norc", Arg.Set noRc,
    "Disable reference counting";
  "--saveadjust", Arg.String (fun dir -> adjustDir := dir; saveAdjust := true),
    "Save generated adjust functions in the specified directory";
  "--fakeadjust", Arg.String (fun dir -> adjustDir := dir; fakeAdjust := true; saveAdjust := true),
    "Create, and save in the specified directory, adjust functions for all types";
  "--warn-typeof-char", Arg.Set warnTypeofChar,
    "Warn when hs_typeof(char) is used - can help detect casts that fool HS_FREE/etc";
  "--customops", Arg.Set_string opsFile,
    "Specify a custom header file";
  "--help", Arg.Unit (fun () -> Arg.usage (align options) ""; exit 0),
    "Show this help message";
  "-help", Arg.Unit (fun () -> Arg.usage (align options) ""; exit 0),
    "Show this help message";

  (* Optimization *)
  "", Arg.Unit (fun () -> ()), "Optimization";
  "--no-hoist", Arg.Set noHoist,
    "Disable hoisting of push/pop operations out of loops";
  "--no-merge", Arg.Set noMerge,
    "Disable removal of redundant pop/push operations";

  (* Things end users usually won't need *)
  "", Arg.Unit (fun () -> ()), "Advanced (for debugging HeapSafe):";
  "--internal-line-nums",
    Arg.Unit (fun _ -> Cil.lineDirectiveStyle := Some Cil.LineComment;
                       Cprint.printLnComment := true),
    "Do not map line numbers back to the original source file";
  "--parsed", Arg.Set_string parsedFile,
    "Output what we parsed"; 
  "--parse-out", Arg.Set_string parseFile,
    "File in which to place parsing results, before preprocessing";
  "--typedebug", Arg.Set typeDebug,
    "Add checks to ensure all pointers are counted (slow, experimental, unsupported)";
  "--nodefer", Arg.Set noDefer,
    "Disable deferred reference counting";
  "--nolocals", Arg.Set noLocals,
    "Disable reference counting on local variables (when possible)";

  (* Internal *)
  "", Arg.Unit (fun () -> ()), "Internal (set by driver):";
  "--home", Arg.Set_string home,
    "HeapSafe home";
  "--out", Arg.Set_string outFile,
    "Output file";
  "--size_t", Arg.Set_string size_t,
    "Explicit specification of size_t type";
]

and align (options: (string * Arg.spec * string) list) =
  (* Get the width of the left column, which contains argument names. *)
  let left =
    try
      List.hd (List.sort (fun a b -> - (compare a b))
                 (List.map (fun (arg, _, _) -> String.length arg) options))
    with Not_found ->
      0
  in
  (* Add extra for left and right margin. *)
  let left = left + 4 in
  (* Now get the width of the description column. *)
  let width = 78 - left in
  (* Helper function to wrap strings. *)
  let rec wrap str =
    if String.length str < width then
      str
    else
      (* Find the point to break the string--first newline or last space. *)
      let break, skip =
        try
          let break = String.rindex_from str width ' ' in
          try
            String.index (String.sub str 0 break) '\n', 1
          with Not_found ->
            break, 1
        with Not_found ->
          width, 0
      in
      (* Split the string and keep wrapping. *)
      let lstr, rstr =
        String.sub str 0 break,
        String.sub str (break + skip) (String.length str - break - skip)
      in
      lstr ^ "\n" ^ String.make left ' ' ^ wrap rstr
  in
  (* Now update all the descriptions. *)
  List.map
    (fun (arg, action, str) ->
       if arg = "" then
         arg, action, "\n" ^ str ^ "\n"
       else
         let pre = String.make (left - String.length arg - 3) ' ' in
         arg, action, pre ^ wrap str)
    options
