(*
 *
 * Copyright (c) 2006,
 *  Jeremy Condit       <jcondit@cs.berkeley.edu>
 *  George C. Necula    <necula@cs.berkeley.edu>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. The names of the contributors may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright (c) 2007 Intel Corporation 
 * All rights reserved. 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 	Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 	Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *     Neither the name of the Intel Corporation nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE INTEL OR ITS
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *)

module F = Frontc
module C = Cil
module P = Pretty
module E = Errormsg

open Rcutils
open Rcinit
open Rc
open Rcprint
open Rcopt
open Cil

let outChannel: out_channel option ref = ref None

let hsPrinter = defaultCilPrinter

exception Done_Processing

let printFile ~(extraPrinting:(out_channel->unit) option) 
              ~(globinit:C.fundec option)
             (f: C.file) (name: string) : unit
  =
  if name <> "" then begin
    try
      let channel = open_out name in
      C.dumpFile hsPrinter channel name f;
      (match globinit with 
        Some g -> C.dumpGlobal hsPrinter
                               channel (C.GFun(g, C.locUnknown))
      | None -> ()); 
      (match extraPrinting with
         Some doit -> doit channel
       | None -> ());
      close_out channel
    with Sys_error _ ->
      E.s (E.error "Error dumping inference results to %s\n" name)
  end

let is_prefix prefix s = 
	prefix = String.sub s 0 (min (String.length prefix) (String.length s))

let isHeapSafeRoot global = match global with
  | GVarDecl(v,_) when is_prefix "rc_adjust" v.vname -> true
  | GFun(f, _) when is_prefix "rc_adjust" f.svar.vname -> true 
  | _ -> Rmtmps.isDefaultRoot global  

let doPragmas global = match global with
  | GPragma (Attr ("hs_norc",_),_) -> Doptions.noRc := true; false
  | _ -> true

let printParsedFile (cabs:Cabs.file) (fname:string) =
  let outchan = open_out fname in
  Cprint.printFile outchan cabs;
  close_out outchan

let parseOneFile (fname: string) : Cabs.file * C.file =
  let cabs, cil = F.parse_with_cabs fname () in
  if !Doptions.parsedFile <> "" then 
        printParsedFile cabs !Doptions.parsedFile;
  Rmtmps.removeUnusedTemps ~isRoot:isHeapSafeRoot cil;
  cabs, cil

let rec processOneFile ((cabs: Cabs.file), (cil: C.file)) =
  try
    cil.globals <- List.filter doPragmas cil.globals;

    printFile ~extraPrinting:None ~globinit:None cil !Doptions.parseFile;

    initRc cil;

    if not !Doptions.noRc then rcProcessFile cil else norcProcessFile cil;
    clearLocals cil;
    
    rcOptimize cil;

    begin
      match !outChannel with
      | None -> ()
      | Some c ->
          (* Tell CIL to put comments around the bounds attributes. *)
          C.print_CIL_Input := false;
	  output_string c ("#include <"^ !Doptions.opsFile^">\n");
          Stats.time "printCIL" 
               (C.dumpFile (!C.printerForMaincil) c !Doptions.outFile) cil
    end;

    if !E.hadErrors then
      E.s (E.error "Error while processing file");
  with Done_Processing -> ()
        
let main () =
  (*print_endline "running hs";*)

(*   Whitetrack.enabled := false; *)

  let usageMsg = "Usage: hscc [options] <source-files>..." in

  (* sm: enabling this by default, since I think usually we
   * want 'cilly' transformations to preserve annotations; I
   * can easily add a command-line flag if someone sometimes
   * wants these suppressed *)
  C.print_CIL_Input := true;

  (* Turn off the implicit casts *)
  C.insertImplicitCasts := false;

  (* Turn off line wrapping. *)
  C.lineLength := 100000;

  (* Suppress truncate warnings. *)
  C.warnTruncate := false;

  Arg.parse (Doptions.align Doptions.options) Ciloptions.recordFile usageMsg;

  if !Doptions.outFile <> "" then begin
    try
      outChannel := Some (open_out !Doptions.outFile)
    with _ ->
      E.s (E.error "Couldn't open file %s\n" !Doptions.outFile)
  end;

  Cil.initCIL ();
  fixSizeT ();
  Ciloptions.fileNames := List.rev !Ciloptions.fileNames;
  (* Use our special printer that prints adjust functions where necessary *)
  C.printerForMaincil := new rcPrinter;
  (* Declare our type-based attributes *)
  Hashtbl.add C.attributeHash "hs_trace" C.AttrType;
  Hashtbl.add C.attributeHash "hs_nofree" C.AttrType;
  Hashtbl.add C.attributeHash "hs_norc" C.AttrType;
  Hashtbl.add C.attributeHash "hs_untyped" C.AttrType;

  let fileName =
    match !Ciloptions.fileNames with
    | [name] -> name
    | [] -> E.s (E.error "No file names provided\n")
    | _ -> E.s (E.error "Too many file names provided (%a)\n"
               (Pretty.docList Pretty.text) !Ciloptions.fileNames)
  in

  let cabs, file = parseOneFile fileName in

  if !E.hadErrors then
    E.s (E.error "Cabs2cil had some errors");

  processOneFile (cabs, file);
  
;;

let failed = ref false 

let cleanup () = 
  if !E.verboseFlag then
    Stats.print !E.logChannel "Timings:\n";
  if !E.logChannel != stderr then 
    close_out (! E.logChannel);  
  match !outChannel with
  | Some c -> close_out c
  | _ -> ()
;;

begin 
  try 
    main () 
  with
  | F.CabsOnly -> (* this is OK *) ()
  | E.Error -> failed := true
end;
cleanup ();
exit (if !failed then 1 else 0)
